/**************************************************************************************************

Copyright (c) 2007-2012   Tsunehiko Kato  and  4D2U Project, National Astronomical Observatory of Japan

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

ȉɒ߂ɏ]A{\tgEFAъ֘Ãt@Ciȉu\tgEFAvj̕擾
邷ׂĂ̐lɑ΂A\tgEFA𖳐ɈƂ𖳏ŋ܂Bɂ́A\tgEFA̕
gpAʁAύXAAfځAЕzATuCZXA/܂͔̔錠Aу\tgEFA
񋟂鑊ɓƂ錠Ɋ܂܂܂B

L̒쌠\і{\A\tgEFÂׂĂ̕܂͏dvȕɋLڂ̂Ƃ
B

\tgEFÁû܂܁vŁAł邩Öقł邩킸A̕ۏ؂Ȃ񋟂܂B
łۏ؂Ƃ́AiA̖ړIւ̓KAьNQɂĂ̕ۏ؂܂݂܂AɌ
̂ł͂܂B ҂܂͒쌠҂́A_sׁAs@sׁA܂͂ȊOł낤ƁA\t
gEFAɋN܂͊֘AA邢̓\tgEFA̎gp܂͂̑̈ɂĐ؂̐A
QȂ̋`ɂĉ̐ӔCȂ̂Ƃ܂B
***************************************************************************************************/
#include  <stdio.h>
#include  "mtk_globular_clusters.h"
#include  "../common/math/random.h"
#include  "../common/graphics/color.h"
#include  "../common/astro/astrometry.h"
#include  "../common/bilingual.h"
#include  "mtk_gl_common.h"
#include  "mtk_font.h"
#include  "mtk_text.h"
#include  "mtk_switches.h"
#include  "mtk_solar_system.h"
#include  "mtk_globular_clusters.h"
#include  "mtk_approx_func.h"
#include  "mtk_our_galaxy_rt.h"
#include  "mtk_colors.h"
#include  "mtk_objects.h"
#include  "mtk_label.h"
#include  "mtk_GC_model.h"
#include  "mtk_calc_alpha.h"




//-------------------------
//  Globular Cluster data
//-------------------------
struct globular_cluster2 : public globular_cluster
{
	vector3f	xyz;		// W
	RGBf		RGB;		// F
	int			draw_type;
	int			sgc_idx;
};


static vector<globular_cluster2>	vGlbc;	// 󐯒c̈ʒu
static vector<vector3f>			GCPT;		// 󐯒c̓_
const int GCPT_N = 500;


//-------------
//  Draw type
//-------------
enum {
	GC_DRAW_TYPE_NORMAL,
	GC_DRAW_TYPE_M13
};



//-----------------------------
//  Special Globular Clusters
//-----------------------------
struct tag_sgc {
	const char*  cat_name;
	const char*  name_E;
	const char*  name_J;
	const char*  key;
	int			 draw_type;
};

const tag_sgc  vSGC[] = {
	{ "NGC 6205",	"M13",			"M13",						"M13",		GC_DRAW_TYPE_M13 },
	{ "NGC 5139",	"Omega Centauri",	"{|IK}{c|}",	"OmegaCentauri", GC_DRAW_TYPE_NORMAL }
};

const int sgc_num = sizeof(vSGC)/sizeof(vSGC[0]);


//------------------
//  Display scales
//------------------
const float GC_scl_min = 1000;
const float GC_scl_max = 12500;





//---------------
//  Load / Init
//---------------
bool
load_globular_cluster_data(const directory& dir)
{
	FILE* fp;
	fp = dir.fopen("globular_cluster.dat", "rb");
	if (fp==NULL)  return false;


	while (true) {
		globular_cluster2 gc;
		
		//--- read data
		int count;
		count = fread(gc.name,		12,			1, fp);
		if (count < 1)	break;
		fread(&gc.eqPos.RAdeg,	sizeof(gc.eqPos.RAdeg),		1, fp);
		fread(&gc.eqPos.DEdeg,	sizeof(gc.eqPos.DEdeg),		1, fp);
		fread(&gc.distance,	sizeof(gc.distance),	1, fp);
		fread(&gc.Rc,		sizeof(gc.Rc),			1, fp);
		fread(&gc.Rh,		sizeof(gc.Rh),			1, fp);
		fread(&gc.B_V,		sizeof(gc.B_V),			1, fp);
		fread(&gc.V_R,		sizeof(gc.V_R),			1, fp);


		//--- Vʏ̈ʒuiPʔaj
		vector3f Cxyz = gc.eqPos.get_dirc_vec();

		//--- SW(pc)
		gc.xyz  = M_R2E * ( gc.distance * Cxyz );

		//--- F
		RGBf rgb;
		double  T = B_V_to_Tc(gc.B_V);
		rgb = correct_RGB( blackbody_color(T) );
		float RGB_max = max(rgb.r, max(rgb.g, rgb.b));
		gc.RGB = rgb / RGB_max;


		//---  check special gc
		int sidx;
		for (sidx=0; sidx<sgc_num; sidx++) {
			if (stricmp(gc.name, vSGC[sidx].cat_name) == 0) break;
		}
		gc.sgc_idx = (sidx==sgc_num)  ?  -1  :  sidx;


		//---  Draw type
		int dtype = GC_DRAW_TYPE_NORMAL;
		if (sidx < sgc_num) {
			dtype = vSGC[sidx].draw_type;
		}
		gc.draw_type = dtype;


		vGlbc.push_back(gc);


		//---  Register object
		int idx = vGlbc.size() - 1;
		char key[256];
		if (sidx < sgc_num)	{ sprintf(key, vSGC[sidx].key); }
		else				{ sprintf(key, "GlobularCluster%d", idx); }
		int obj_idx = register_object( mtk_object(OBJ_TYPE_GLOBULAR_CLUSTER, idx, 0.0f, true), key);


		//---  Register label
		if (sidx < sgc_num) {
			const tag_sgc&  sgc = vSGC[sidx];
			if (_stricmp(gc.name, sgc.cat_name) == 0) {
				sprintf(key, "%sLabel", sgc.key);
				add_label(sgc.name_J, sgc.name_E, obj_idx, MTK_COLOR_NAME_OBJECT, 1.0f, get_object_name_size(), 5*gc.Rh, 10*D2R, true, key);
			}
		}
	}

	return  true;
}


void
init_globular_cluster_points()
{
	for (int i=0; i<GCPT_N; ++i) {
		float phi = rndUni(0,2*pi);
		float mu  = rndUni(-1,1);
		float x,y,z,rho, R;

		R = rndGauss(0, 1);

		z   = R*mu;
		rho = R*sqrt(1-mu*mu);
		x   = rho*cos(phi);
		y   = rho*sin(phi);
		
		GCPT.push_back(vector3f(x,y,z));
	}
}



//-------
//  Pos
//-------
vector3d
get_globular_cluster_position(int idx)
{
	return  vGlbc[idx].xyz;
}



//---------
//  Scale
//---------
double
get_scale_globular_cluster(int idx)
{
	const globular_cluster2& gc = vGlbc[idx];
	return  30.0 * gc.Rh;
}


//---------
//  Label
//---------
const double gc_lc1 = 1000;
const double gc_lc2 = 3000;

bool
is_globular_cluster_label_visible(const mtk_draw_struct& ds, int idx)
{
	if ( !(swGalaxy & GLXY_GLOB_CLSTR && test_disp_switch(DISPSW_PLANET, DISPM_NAME)) )  return false;
	if ( ds.cnorm < gc_lc1 )  return  false;

	return  true;
}


float
get_globular_cluster_label_alpha(const mtk_draw_struct& ds, int idx)
{
	const double lc = ds.cnorm;
	if (lc < gc_lc2) {
		return  (lc - gc_lc1) / (gc_lc2 - gc_lc1);
	}


	const vector3d   pos = get_globular_cluster_position(idx);
	const vector3d  cpos = ds.camera.get_pos();
	const vector3d  rpos = pos - cpos;
	const vector3d  gpos = pos - get_our_galaxy_position();
	float  Rv  = 0.5f * gpos.norm();
	float  Dth = 2.0f * Rv / rpos.norm();
	return  calc_alpha_by_apparent_diameter( Dth );
}




//------------------
//  Landing radius
//------------------
double
get_landing_radius_globular_cluster(int idx)
{
	return vGlbc[idx].Rh;
}




//----------------
//  Prepare draw
//----------------
bool
prepare_draw_globular_cluster(const mtk_draw_struct& ds, int idx, double* pRb, double* pRh)
{
	if ( !(swGalaxy & GLXY_GLOB_CLSTR) )  return false;
	
	
	//***
	if (ds.cnorm < GC_scl_min)  return false;


	const globular_cluster2& gc = vGlbc[idx];
	const float rate  = 10.0;				// Rh ̉{󐯒c̑傫ƌȂ

	*pRb = rate * gc.Rh;
	*pRh = gc.Rh;

	return  true;
}



//--------
//  Draw
//--------
void
draw_a_globular_cluster(const mtk_draw_struct& ds, int idx, const vector3d& rpos, const double& scl)
{
	const globular_cluster2& gc = vGlbc[idx];
	const double scale = ds.cnorm;
	float af = (scale > GC_scl_max)  ?  1  :  (scale-GC_scl_min) / (GC_scl_max-GC_scl_min);
	if (af < 0.01f)  return;


	if (gc.draw_type == GC_DRAW_TYPE_M13) {
		draw_GC_model(ds, idx, rpos, scl, 1.0);
		return;
	}


	const vector3d camera_pos = ds.camera.get_pos();


	//-- optical depth
	float tauR = calc_galactic_optical_depth(camera_pos, gc.xyz, 3);
	if (tauR > 3)  return;

	float _r, _g, _b;
	_r = gc.RGB.r * exp_d(tauR);
	_g = gc.RGB.g * exp_d(kpG_kpR*tauR);
	_b = gc.RGB.b * exp_d(kpB_kpR*tauR);



	glEnable(GL_BLEND);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	glDepthMask(GL_FALSE);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE);


	const float c = af * 0.8f;
	glColor3f(c*_r, c*_g, c*_b);

	const float fac = scl * gc.Rh;



	//---  JWn̎
	vector3d  eF, eU, eR;
	ds.camera.get_axes(eF, eU, eR);
	const float a = 0.4 / get_screen_z() * get_screen_w() / 180;
	const vector3f eUf = vector3f( eU );



	glPushMatrix();

		if (true) {
			// QUAD

			glBegin(GL_QUADS);
			for (int i=0; i<GCPT.size(); i++) {
				vector3f  rr = fac * GCPT[i] + rpos;

				//---  lpWn̎
				vector3f  eF2, eU2, eR2;
				float R2, R;
				R2 = norm2(rr);
				if (R2 < 1e-100)  continue;
				
				R = sqrt(R2);
				eF2 = rr / R;
				eR2 = normalize( cross_product(eF2, eUf) );
				eU2 = cross_product( eR2, eF2 );

				float sz;
				sz = a * R;

				vector3f  r;
				r = rr + sz*eR2;
				glVertex3fv( r.get_pointer() );
				r = rr + sz*eU2;
				glVertex3fv( r.get_pointer() );
				r = rr - sz*eR2;
				glVertex3fv( r.get_pointer() );
				r = rr - sz*eU2;
				glVertex3fv( r.get_pointer() );
			}
			glEnd();
		} else {
			glBegin(GL_POINTS);
				for (int j=0; j<GCPT.size(); j++) {
					vector3f xyz = fac * GCPT[j];
					glVertex3fv(xyz.get_pointer());
				}
			glEnd();
		}
	glPopMatrix();
}



//--------
//  Info
//--------
void
make_object_info_globular_cluster(const mtk_draw_struct& ds, int idx, std::vector<std::string>& text)
{
	const globular_cluster2&  gc = vGlbc[idx];
	int sgc_idx = gc.sgc_idx;


	if (ds.lang == LANG_ENG) {
		if (sgc_idx != -1) {
			text.push_back( vSGC[sgc_idx].name_E );
		}
		text.push_back( gc.name );
		text.push_back( "Type: Globular Cluster" );
	} else {
		if (sgc_idx != -1) {
			text.push_back( vSGC[sgc_idx].name_J );
		}
		text.push_back( gc.name );
		text.push_back( "ށF󐯒c" );
	}
}





//------
//  Rh
//------
double
get_globular_cluster_Rh(int idx)
{
	return  vGlbc[idx].Rh;
}
